package main

import (
	"bytes"
	"crypto/sha256"
	"encoding/json"
	"fmt"
	"io/ioutil"
	"net/http"
	"strings"

	"github.com/go-icap/icap"
)

// Config represents the Server config
type Config struct {
	Methods         string
	Allow204        bool
	Preview         int
	TransferPreview string
	ListenAddr      string
	ListenPort      int
	Path            string
	ISTag           string
}

// Server represents the icap server
type Server struct {
	Config
}

func main() {
	s := NewServer(Config{
		Methods:         getEnv("ICAP_METHODS", "REQMOD"),
		Allow204:        getEnvB("ICAP_ALLOW204", "true"),
		Preview:         getEnvI("ICAP_PREVIEW_SIZE", "4096"),
		TransferPreview: getEnv("ICAP_PREVIEW_FILTER", "*"),
		ListenAddr:      getEnv("ICAP_LISTEN_ADDR", "0.0.0.0"),
		ListenPort:      getEnvI("ICAP_LISTEN_PORT", "1344"),
		Path:            getEnv("ICAP_PATH", "/reqmod"),
	})
	s.Start()
}

// NewServer creates and initialize a new Server object
func NewServer(conf Config) Server {
	s := Server{
		Config: conf,
	}

	// Calculate ISTag
	confJSON, err := json.Marshal(conf)
	if err != nil {
		panic("Error marshaling config")
	}
	s.ISTag = fmt.Sprintf("%x", sha256.Sum256(confJSON))
	icap.HandleFunc(conf.Path, s.reqHandler)
	return s
}

// Start ICAP server
func (s *Server) Start() {
	addr := fmt.Sprintf("%s:%d", s.ListenAddr, s.ListenPort)
	icap.ListenAndServe(addr, icap.HandlerFunc(s.reqHandler))
}

func (s *Server) reqHandler(resp icap.ResponseWriter, req *icap.Request) {
	h := resp.Header()
	h.Set("ISTag", s.ISTag)
	h.Set("Service", "ICAP Test")

	switch req.Method {
	case "OPTIONS":
		s.options(resp)
	case "REQMOD":
		s.reqmod(resp, req)
	default:
		resp.WriteHeader(405, nil, false)
		fmt.Println("Invalid request method")
	}
}

func (s *Server) options(resp icap.ResponseWriter) {
	h := resp.Header()
	h.Set("Methods", s.Methods)
	if s.Allow204 {
		h.Set("Allow", "204")
	}
	if s.Preview > 0 {
		h.Set("Preview", fmt.Sprintf("%d", s.Preview))
		h.Set("Transfer-Preview", s.TransferPreview)
	}
	resp.WriteHeader(200, nil, false)
}

func (s *Server) response(body string, req *http.Request) *http.Response {
	return &http.Response{
		Status:        "403 Forbidden",
		StatusCode:    403,
		Proto:         "HTTP/1.1",
		ProtoMajor:    1,
		ProtoMinor:    1,
		Body:          ioutil.NopCloser(bytes.NewBufferString(body)),
		ContentLength: int64(len(body)),
		Request:       req,
		Header:        make(http.Header, 0),
	}
}

func (s *Server) reqmod(resp icap.ResponseWriter, req *icap.Request) {

	buffer, err := ioutil.ReadAll(req.Request.Body)
	if err != nil {
		panic(err)
	}

	// Check for a bad filename
	if strings.Contains(req.Request.URL.RequestURI(), "failme") {
		resp.WriteHeader(200, s.response("bad filename", req.Request), true)
		resp.Write([]byte("bad filename"))
		return
	}

	// Check for a bad preview
	if strings.Contains(string(req.Preview), "failme") {
		resp.WriteHeader(200, s.response("bad preview", req.Request), true)
		resp.Write([]byte("bad preview"))
		return
	}

	// Check for a bad body
	if strings.Contains(string(buffer), "failme") {
		resp.WriteHeader(200, s.response("bad body", req.Request), true)
		resp.Write([]byte("bad body"))
		return
	}

	resp.WriteHeader(204, nil, false)
}
